"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = require("child_process");
const util_1 = require("util");
const os_1 = require("os");
const fs_1 = require("fs");
const path_1 = require("path");
const lodash_1 = require("lodash");
const file_1 = require("./file");
const execA = util_1.promisify(child_process_1.exec);
const writeFileA = util_1.promisify(fs_1.writeFile);
const copyFileA = util_1.promisify(fs_1.copyFile);
const accessA = util_1.promisify(fs_1.access);
class Compiler {
    constructor(compiler) {
        this.compiler = compiler;
        this.code = [];
        this.includes = ['dbg.h'];
        if (this.compiler === 'gcc' || this.compiler === 'clang')
            this.langExt = 'c';
        else
            this.langExt = 'cxx';
    }
    static wrapCodeWith(code, wrapper = 'value_of') {
        code = lodash_1.trim(code, ';');
        return `${wrapper}(${code});`;
    }
    static wrapInclude(module) {
        return `#include "${module}"`;
    }
    static addSemicolon(code) {
        code = lodash_1.trim(code, ';');
        return `${code};`;
    }
    static concatCode(code, includes) {
        let resCode = '';
        let resInclude = '';
        for (const i of includes) {
            resInclude = resInclude.concat(Compiler.wrapInclude(i)) + '\n';
        }
        for (const i of code) {
            resCode = resCode.concat(i) + '\n';
        }
        return `${resInclude}\nint main () {\n${resCode}\n}`;
    }
    static async isExist(path) {
        try {
            await accessA(path_1.join(os_1.tmpdir(), 'dbg.h'));
            return true;
        }
        catch (e) {
            return false;
        }
    }
    async basicCompile(code, includes) {
        let resCode = Compiler.concatCode(code, includes);
        const filepath = path_1.join(os_1.tmpdir(), `repl.${this.langExt}`);
        const headerpath = path_1.join(os_1.tmpdir(), 'dbg.h');
        const execPath = path_1.join(os_1.tmpdir(), file_1.getOSExecName());
        await writeFileA(filepath, resCode);
        let exec_res;
        await copyFileA(await file_1.getLocalPath(path_1.join('template', 'dbg.h')), headerpath);
        let cxx17 = '';
        if (this.langExt == 'cxx') {
            cxx17 = '-std=c++17';
        }
        try {
            const execCommand = `${this.compiler} -w ${filepath} ${cxx17} -fdiagnostics-color -o ${execPath}`;
            exec_res = await execA(execCommand, { windowsHide: true });
            return {
                success: true,
                output: exec_res.stderr
            };
        }
        catch (e) {
            return {
                success: false,
                output: e.stderr
            };
        }
    }
    static isCommand(code) {
        code = lodash_1.trim(code);
        if (code === '')
            return false;
        return code[0] === ':';
    }
    async processInput(line) {
        return Compiler.isCommand(line) ? this.command(line) : this.compile(line);
    }
    async command(cmd) {
        try {
            const oprandArr = cmd.split(' ');
            let oprand = '';
            for (let i = 1; i < oprandArr.length; ++i) {
                oprand += oprandArr[i] + (i == oprandArr.length - 1 ? '' : ' ');
            }
            let tmp = [...this.code];
            let r;
            switch (oprandArr[0]) {
                case ':b':
                case ':bin':
                    tmp.push(Compiler.wrapCodeWith(oprand, 'print_bytes'));
                    r = await this.basicCompile(tmp, this.includes);
                    return r;
                case ':x':
                case ':hex':
                    tmp.push(Compiler.wrapCodeWith(oprand, 'print_hex'));
                    r = await this.basicCompile(tmp, this.includes);
                    return r;
                case ':h':
                case ':help':
                    return {
                        success: true,
                        output: ''
                    };
                case ':m':
                case ':module':
                    let tmpInclude = [...this.includes];
                    tmpInclude.push(oprand);
                    let compileRes = await this.basicCompile(this.code, tmpInclude);
                    if (compileRes.success) {
                        this.includes.push(oprand);
                    }
                    return {
                        success: compileRes.success,
                        output: compileRes.success ? `Load ${oprand} success\n` : `Fail to load ${oprand}\n`
                    };
                case ':t':
                case ':type':
                    tmp.push(Compiler.wrapCodeWith(oprand, 'type_of'));
                    r = await this.basicCompile(tmp, this.includes);
                    return r;
                case ':q':
                case ':quit':
                    process.exit(0);
            }
        }
        catch (e) {
            return {
                success: false,
                output: e.message
            };
        }
        return {
            success: false,
            output: 'match command failed\n'
        };
    }
    async compile(code) {
        code = Compiler.addSemicolon(code);
        let tmp = [...this.code];
        tmp.push(Compiler.wrapCodeWith(code));
        let r = await this.basicCompile(tmp, this.includes);
        if (!r.success) {
            tmp.pop();
            tmp.push(code);
            r = await this.basicCompile(tmp, this.includes);
            if (r.success) {
                this.code.push(code);
            }
        }
        else {
            this.code.push(code);
        }
        return r;
    }
    async execute() {
        const execPath = path_1.join(os_1.tmpdir(), file_1.getOSExecName());
        const res = await execA(execPath, { windowsHide: true });
        return res.stdout;
    }
}
exports.default = Compiler;
